#!/usr/bin/fades --system-site-packages

import enum
import datetime
import sys
from decimal import Decimal

import pyoo  # fades

CONNECT_MSG = """
Need to start Libreffice! Use a command like:
    libreoffice -accept="socket,host=localhost,port=2002;urp;" &
"""

# all useful rows should be less than this
MAX_ROW_DATA = 3000

# epoch
BASE_DATE = datetime.date(1899, 12, 30)

# to round to 4 decimal points
DECIMAL_4_DIGITS = Decimal(".0001")


class Record:
    """A typical record with expense info."""
    def __init__(self, triplet):
        date_int, self.description, amount_float = triplet
        self.date = datetime.timedelta(days=date_int) + BASE_DATE
        self.amount = Decimal.from_float(amount_float).quantize(DECIMAL_4_DIGITS)

    def __str__(self):
        return "[{:%Y-%m-%d}] {}: {}".format(self.date, self.description, self.amount)


if len(sys.argv) != 3:
    print("Usage: expenses_for_accountant.py <year> <path_to_ods>")
    exit()

try:
    desktop = pyoo.Desktop('localhost', 2002)
except IOError:
    print(CONNECT_MSG)
    exit()
YEAR = int(sys.argv[1])
doc = desktop.open_spreadsheet(sys.argv[2])


def get_records(sheet, init_row, init_col):
    """Get the records from a pack."""
    pack = sheet[init_row:MAX_ROW_DATA, init_col:init_col + 3].values
    pack = iter(pack)

    # consume the header
    triplet = next(pack)
    assert triplet[1] != ''  # the title
    triplet = next(pack)
    assert triplet[0] == ''
    assert triplet[2] != ''  # the total of the column
    triplet = next(pack)
    assert triplet == ('', '', ''), triplet  # the separator

    # read initial chunk (the visible one where we normally operate)
    initial_chunk = []
    for triplet in pack:
        if triplet[0] == '':
            # end of first chunk
            break
        initial_chunk.append(triplet)

    # consume the middle intermediate space
    for triplet in pack:
        if triplet[0] != '':
            break

    # grab the historic data
    historic = [triplet]
    for triplet in pack:
        if triplet[0] == '':
            break
        historic.append(triplet)
    else:
        raise ValueError("End of data achieved without emptiness, MAX_ROW_DATA not enough")

    # join and process
    all_data = [Record(triplet) for triplet in historic + initial_chunk]
    return all_data


def process_pack(sheet, init_row, init_col, title, expense_type):
    """Process a pack of three columns."""
    print("Processing", sheet.name, repr(title))
    raw_records = get_records(sheet, init_row, init_col)
    raw_records = [r for r in raw_records if r.date.year == YEAR]

    # validate date ordering
    if not raw_records:
        print("WARNING: no records for", sheet, init_row, init_col, title)
        return title, "-", 0
    prev = raw_records[0]
    for record in raw_records[1:]:
        if record.date < prev.date:
            print("    bad date order!", prev, record)
        prev = record

    # get only negative values
    records = []
    for record in raw_records:
        if record.amount > 0:
            desc = record.description.lower()
            if 'dev' not in desc:
                if desc != 'mes' and desc != 'viáticos' and not desc.startswith("venta"):
                    print("     positive??", record)
                continue

        # convert to "expensed"
        record.amount *= -1

        records.append(record)

    # group by month
    grouped_by_month = {}
    for record in records:
        grouped_by_month.setdefault("{:%Y-%m}".format(record.date), []).append(record)

    # if expense is just a group, add and show
    if expense_type == ExpenseType.group:
        for month, records in sorted(grouped_by_month.items()):
            yield title, month, sum(r.amount for r in records)
        return

    # need to get into the detail
    grouped_by_detail = {}
    for month, records in grouped_by_month.items():
        for record in records:
            key = (month, record.description.split()[0])
            grouped_by_detail.setdefault(key, []).append(record)

    for (month, detail), records in sorted(grouped_by_detail.items()):
        composed_title = title + ' ' + detail
        total = sum(r.amount for r in records)
        yield composed_title, month, total


ExpenseType = enum.Enum("ExpenseType", "group detail")
COLLECT_DATA = {
    "Personas": (0, {
        "Facundo": ("Gastos generales Facundo", ExpenseType.group, 50),
        "Moni": ("Gastos generales Mónica", ExpenseType.group, 50),
        "Felipe": ("Gastos generales Felipe", ExpenseType.group, 50),
        "Malena": ("Gastos generales Malena", ExpenseType.group, 50),
        "Vacaciones": ("Vacaciones", ExpenseType.group, 50),
        "Actividades": ("Actividad:", ExpenseType.detail, 15),
    }),
    "Hogar": (0, {
        "Casa": ("Día a día de la casa", ExpenseType.group, 50),
        "Ahorro general": ("Caja chica de la casa", ExpenseType.group, 80),
        "Arreglos/mejoras casa": ("Arreglos/mejoras de casa", ExpenseType.group, 60),
        "Empleados": ("Empleados:", ExpenseType.detail, 5),
        "Salud": ("Salud:", ExpenseType.detail, 6),
    }),
    "Servicios": (0, {
        "Agua": ("Servicios: Agua", ExpenseType.group, 0),
        "Electricidad": ("Servicios: Electricidad", ExpenseType.group, 0),
        "Gas": ("Servicios: Gas", ExpenseType.group, 0),
        "Teléfono": ("Servicios: Teléfono", ExpenseType.group, 0),
        "Municipalidad": ("Servicios: Municipalidad", ExpenseType.group, 0),
        "ARBA": ("Servicios: ARBA", ExpenseType.group, 0),
        "Cable": ("Servicios: Cable", ExpenseType.group, 50),
        "Internet": ("Servicios: Internet", ExpenseType.group, 0),
    }),
    "Auto": (0, {
        "Estac.": ("Auto: Estacionamientos", ExpenseType.group, 30),
        "Seguro": ("Auto: Seguro", ExpenseType.group, 0),
        "Patente": ("Auto: Patente", ExpenseType.group, 0),
        "Mantenimiento": ("Auto: Mantenimiento", ExpenseType.group, 20),
        "Nafta": ("Auto: Nafta", ExpenseType.group, 5),
        "Peajes": ("Auto: Peajes", ExpenseType.group, 0),
    }),
    "Trabajo": (8, {
        "Costos Laburo": ("Gastos operativos trabajo", ExpenseType.group, 0),
    }),
}

all_yms = set()
data = {}
for sheet in doc.sheets:
    try:
        init_row, collect_info = COLLECT_DATA[sheet.name]
    except KeyError:
        continue

    # walk on the titles row, finding useful pack of columns
    cells = sheet[init_row, 0:100].values
    for column, cell in enumerate(cells):
        if cell in collect_info:
            title, expense_type, blackporc = collect_info[cell]
            pivot = column - 1
            results = process_pack(sheet, init_row, pivot, title, expense_type)
            for title, yearmonth, amount in results:
                all_yms.add(yearmonth)
                amount *= (100 - Decimal(blackporc)) / 100
                data.setdefault(title, {})[yearmonth] = amount

print("\nRESULTS:")
all_yms = sorted(all_yms)
print("\t" + "\t".join(all_yms))
for title, per_ym in sorted(data.items()):
    print(title + "\t" + "\t".join(
        "{:.2f}".format(per_ym[ym]) if ym in per_ym else "" for ym in all_yms))
